import gymnasium as gym
import torch
import numpy as np
import matplotlib.pyplot as plt

# modify the dropout rate here
DROPOUT_RATE = 0.5

# the policy class
class Policy(torch.nn.Module):
    def __init__(self, input_size=4, output_size=2):
        super(Policy, self).__init__()
        self.fc1 = torch.nn.Linear(input_size, 128)
        self.relu = torch.nn.ReLU()
        self.dropout = torch.nn.Dropout(DROPOUT_RATE)
        self.fc2 = torch.nn.Linear(128, output_size)
        self.softmax = torch.nn.Softmax(dim=0)
    
    def forward(self, x):
        x = self.fc1(x)
        x = self.relu(x)
        x = self.dropout(x)
        x = self.fc2(x)
        x = self.softmax(x)
        return x


def main():

    # load policy and optimizer
    policy = Policy()
    optimizer = torch.optim.Adam(policy.parameters(), lr=5e-3)

    # create the environment
    env = gym.make("CartPole-v1")

    gamma = 0.99
    # track total loss and reward
    total_reward = []
    total_loss = []
    epochs = 500

    # max possible steps in the envorinment
    max_steps = env.spec.max_episode_steps


    for ep in range(epochs):
        print(ep)
        observation = env.reset()[0]

        # allocate enough spaces for the rewards and log_probs
        rewards = torch.zeros(max_steps)
        log_probs = torch.zeros(max_steps)
        for step in range(max_steps):

            action_probs = policy(torch.from_numpy(observation).float())

            action = torch.distributions.Categorical(action_probs).sample()
            observation, reward, terminated, truncated, info = env.step(action.numpy())

            # set rewards and log_probs
            rewards[step] = reward
            log_probs[step] = torch.log(action_probs[action])

            if terminated or truncated:
                break
        
        # calculate discounted rewards in reverse
        R = 0
        returns = []
        for r in reversed(rewards[:step+1]):
            R = r + gamma * R
            returns.insert(0, R)
        returns_tensor = torch.tensor(returns)
        eps = 1e-10
        # normalize the returns
        normalized_rewards = (returns_tensor - returns_tensor.mean()) / (returns_tensor.std() + eps)
        log_probs_truncated = log_probs[:step+1]
        loss = torch.sum(-log_probs_truncated * normalized_rewards)
        total_reward.append(sum(rewards[:step+1]))
        # optimize
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        total_loss.append(loss.detach().numpy())

    # save the model weights
    torch.save(policy.state_dict(), "reinforce_cartpole.pth")

    env.close()

    # plot the rewards and the loss side by side
    fig, ax = plt.subplots(1,2)
    ax[0].plot(total_reward)
    ax[1].plot(total_loss)
    ax[0].set_title("Rewards")
    ax[1].set_title("Loss")
    plt.savefig(f"reinforce_cartpole_dr_{DROPOUT_RATE}.png")




if __name__ == "__main__":
    main()